/* (C) 1999-2001 Paul `Rusty' Russell
 * (C) 2002-2004 Netfilter Core Team <coreteam@netfilter.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
/* Modification History
 * --------------------
 * 03Feb10,sp  SPR-16927 DUT's WAN Address accessing issue fix
 * 23Nov09,sg  cleaning sip conntrack entry with NETDEV_GENERIC event
 * 19Nov09,sg  cleaning sip conntrack entries when get an NETDEV_UP
 * 28aug09,ma  Changes to allow strict one to one IB NAT. SPR 13349
 */

/* Everything about the rules for NAT. */
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/kmod.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <linux/inetdevice.h>
#include <net/checksum.h>
#include <net/route.h>
#include <linux/bitops.h>

#define ASSERT_READ_LOCK(x) MUST_BE_READ_LOCKED(&ip_nat_lock)
#define ASSERT_WRITE_LOCK(x) MUST_BE_WRITE_LOCKED(&ip_nat_lock)

#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_core.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/listhelp.h>
#include <linux/netfilter_ipv4/lockhelp.h>

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#define TEAMF1_CHANGES

#define NAT_VALID_HOOKS ((1<<NF_IP_PRE_ROUTING) | (1<<NF_IP_POST_ROUTING) | (1<<NF_IP_LOCAL_OUT))

#ifdef TEAMF1_CHANGES
#define SIP_PORT 0x13C4
#define TF1_DBG 1000
/* Lock protects snat region inside conntrack */
static DECLARE_RWLOCK(snat_lock);
#define TEAMF1_OTO_IB_NAT_CHANGES
#endif /* TEAMF1_CHANGES */

/* Enable Strict One to One Inbound Nat */
int  strictDnat = 1;
/* Standard entry. */
/*struct ipt_standard
{
	struct ipt_entry entry;
	struct ipt_standard_target target;
};

struct ipt_error_target
{
	struct ipt_entry_target target;
	char errorname[IPT_FUNCTION_MAXNAMELEN];
};

struct ipt_error
{
	struct ipt_entry entry;
	struct ipt_error_target target;
};*/

static struct
{
	struct ipt_replace repl;
	struct ipt_standard entries[3];
	struct ipt_error term;
} nat_initial_table __initdata
= { { "nat", NAT_VALID_HOOKS, 4,
      sizeof(struct ipt_standard) * 3 + sizeof(struct ipt_error),
      { [NF_IP_PRE_ROUTING] = 0,
	[NF_IP_POST_ROUTING] = sizeof(struct ipt_standard),
	[NF_IP_LOCAL_OUT] = sizeof(struct ipt_standard) * 2 },
      { [NF_IP_PRE_ROUTING] = 0,
	[NF_IP_POST_ROUTING] = sizeof(struct ipt_standard),
	[NF_IP_LOCAL_OUT] = sizeof(struct ipt_standard) * 2 },
      0, NULL, { } },
    {
	    /* PRE_ROUTING */
	    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
		0,
		sizeof(struct ipt_entry),
		sizeof(struct ipt_standard),
		0, { 0, 0 }, { } },
	      { { { { IPT_ALIGN(sizeof(struct ipt_standard_target)), "" } }, { } },
		-NF_ACCEPT - 1 } },
	    /* POST_ROUTING */
	    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
		0,
		sizeof(struct ipt_entry),
		sizeof(struct ipt_standard),
		0, { 0, 0 }, { } },
	      { { { { IPT_ALIGN(sizeof(struct ipt_standard_target)), "" } }, { } },
		-NF_ACCEPT - 1 } },
	    /* LOCAL_OUT */
	    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
		0,
		sizeof(struct ipt_entry),
		sizeof(struct ipt_standard),
		0, { 0, 0 }, { } },
	      { { { { IPT_ALIGN(sizeof(struct ipt_standard_target)), "" } }, { } },
		-NF_ACCEPT - 1 } }
    },
    /* ERROR */
    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
	0,
	sizeof(struct ipt_entry),
	sizeof(struct ipt_error),
	0, { 0, 0 }, { } },
      { { { { IPT_ALIGN(sizeof(struct ipt_error_target)), IPT_ERROR_TARGET } },
	  { } },
	"ERROR"
      }
    }
};

static struct ipt_table nat_table = {
	.name		= "nat",
	.table		= &nat_initial_table.repl,
	.valid_hooks	= NAT_VALID_HOOKS,
	.lock		= RW_LOCK_UNLOCKED,
	.me		= THIS_MODULE,
	.af		= AF_INET,
};

/* Source NAT */
static unsigned int ipt_snat_target(struct sk_buff **pskb,
				    const struct net_device *in,
				    const struct net_device *out,
				    unsigned int hooknum,
				    const void *targinfo,
				    void *userinfo)
{
	struct ip_conntrack *ct;
	enum ip_conntrack_info ctinfo;
	const struct ip_nat_multi_range_compat *mr = targinfo;

	IP_NF_ASSERT(hooknum == NF_IP_POST_ROUTING);

	ct = ip_conntrack_get(*pskb, &ctinfo);

	/* Connection must be valid and new. */
	IP_NF_ASSERT(ct && (ctinfo == IP_CT_NEW || ctinfo == IP_CT_RELATED
	                    || ctinfo == IP_CT_RELATED + IP_CT_IS_REPLY));
	IP_NF_ASSERT(out);
#ifdef TEAMF1_CHANGES    
	WRITE_LOCK(&snat_lock);
	ct->nat.snat_index = out->ifindex;
	WRITE_UNLOCK(&snat_lock);
#endif /* TEAMF1_CHANGES */

	return ip_nat_setup_info(ct, &mr->range[0], hooknum);
}

/* Before 2.6.11 we did implicit source NAT if required. Warn about change. */
static void warn_if_extra_mangle(u32 dstip, u32 srcip)
{
	static int warned = 0;
	struct flowi fl = { .nl_u = { .ip4_u = { .daddr = dstip } } };
	struct rtable *rt;

	if (ip_route_output_key(&rt, &fl) != 0)
		return;

	if (rt->rt_src != srcip && !warned) {
		printk("NAT: no longer support implicit source local NAT\n");
		printk("NAT: packet src %u.%u.%u.%u -> dst %u.%u.%u.%u\n",
		       NIPQUAD(srcip), NIPQUAD(dstip));
		warned = 1;
	}
	ip_rt_put(rt);
}

static unsigned int ipt_dnat_target(struct sk_buff **pskb,
				    const struct net_device *in,
				    const struct net_device *out,
				    unsigned int hooknum,
				    const void *targinfo,
				    void *userinfo)
{
	struct ip_conntrack *ct;
	enum ip_conntrack_info ctinfo;
	const struct ip_nat_multi_range_compat *mr = targinfo;
    unsigned int ret = 0;

	IP_NF_ASSERT(hooknum == NF_IP_PRE_ROUTING
		     || hooknum == NF_IP_LOCAL_OUT);

	ct = ip_conntrack_get(*pskb, &ctinfo);

	/* Connection must be valid and new. */
	IP_NF_ASSERT(ct && (ctinfo == IP_CT_NEW || ctinfo == IP_CT_RELATED));

	if (hooknum == NF_IP_LOCAL_OUT
	    && mr->range[0].flags & IP_NAT_RANGE_MAP_IPS)
		warn_if_extra_mangle((*pskb)->nh.iph->daddr,
				     mr->range[0].min_ip);

#ifdef TEAMF1_CHANGES
        /*
         * This mark is set to identify all the dnat'd packets, so that
         * the dnat'ed packets can be bypassed in the filter table with
         * out much examination. 
         */
        ct->mark = 0x2222;
#endif /* TEAMF1_CHANGES */


#ifdef TEAMF1_OTO_IB_NAT_CHANGES
        if (strictDnat)
        if(mr->otoNatOrgDstMinIp && mr->otoNatOrgDstMaxIp) {
            /* Get the source address and calculate the IP to DNAT to */
            u_int32_t dstIp;
            u_int32_t offset;
            dstIp = ct->tuplehash[0].tuple.dst.ip;

            /* We do not know how we reached here, its fate be decided
             * un-modified.
             */
            if (dstIp > mr->otoNatOrgDstMaxIp ||
                dstIp < mr->otoNatOrgDstMinIp)
                return IPT_CONTINUE;  

            offset = dstIp - mr->otoNatOrgDstMinIp;

            struct ip_nat_multi_range_compat *strictRangeCompat = NULL;

            strictRangeCompat = kmalloc(sizeof
                                       (struct ip_nat_multi_range_compat),
                                        GFP_KERNEL);
            if (!strictRangeCompat) {
                printk(KERN_WARNING "%s:No memory !\n", __FUNCTION__);
                return 0;
            }
            memcpy(strictRangeCompat, mr, sizeof(struct
                                          ip_nat_multi_range_compat));
            /* We need to reset max_ip here, otherwise ip_nat_setup_info
             * randomly choose an IP between min_ip and max_ip
             */

            strictRangeCompat->range[0].min_ip += offset;
            strictRangeCompat->range[0].max_ip = 
                strictRangeCompat->range[0].min_ip;
            

            DEBUGP("%u.%u.%u.%u OTO IB NAT to %u.%u.%u.%u..",
                   NIPQUAD(dstIp),
                   NIPQUAD(strictRangeCompat->range[0].min_ip));
            /* TODO: OTO IB NAT on range ports */
	        ret = ip_nat_setup_info(ct, &strictRangeCompat->range[0], hooknum);
            kfree(strictRangeCompat);
            return ret;
        }
#endif
defaultNat:
	return ip_nat_setup_info(ct, &mr->range[0], hooknum);
}

static int ipt_snat_checkentry(const char *tablename,
			       const void *entry,
			       void *targinfo,
			       unsigned int targinfosize,
			       unsigned int hook_mask)
{
	struct ip_nat_multi_range_compat *mr = targinfo;

	/* Must be a valid range */
	if (mr->rangesize != 1) {
		printk("SNAT: multiple ranges no longer supported\n");
		return 0;
	}

	if (targinfosize != IPT_ALIGN(sizeof(struct ip_nat_multi_range_compat))) {
		DEBUGP("SNAT: Target size %u wrong for %u ranges\n",
		       targinfosize, mr->rangesize);
		return 0;
	}

	/* Only allow these for NAT. */
	if (strcmp(tablename, "nat") != 0) {
		DEBUGP("SNAT: wrong table %s\n", tablename);
		return 0;
	}

	if (hook_mask & ~(1 << NF_IP_POST_ROUTING)) {
		DEBUGP("SNAT: hook mask 0x%x bad\n", hook_mask);
		return 0;
	}
	return 1;
}

static int ipt_dnat_checkentry(const char *tablename,
			       const void *entry,
			       void *targinfo,
			       unsigned int targinfosize,
			       unsigned int hook_mask)
{
	struct ip_nat_multi_range_compat *mr = targinfo;

	/* Must be a valid range */
	if (mr->rangesize != 1) {
		printk(KERN_DEBUG "DNAT: multiple ranges no longer supported\n");
		return 0;
	}

	if (targinfosize != IPT_ALIGN(sizeof(struct ip_nat_multi_range_compat))) {
		printk(KERN_DEBUG "DNAT: Target size %u wrong for %u ranges\n",
		       targinfosize, mr->rangesize);
		return 0;
	}

	/* Only allow these for NAT. */
	if (strcmp(tablename, "nat") != 0) {
		printk(KERN_DEBUG "DNAT: wrong table %s\n", tablename);
		return 0;
	}

	if (hook_mask & ~((1 << NF_IP_PRE_ROUTING) | (1 << NF_IP_LOCAL_OUT))) {
		printk(KERN_DEBUG "DNAT: hook mask 0x%x bad\n", hook_mask);
		return 0;
	}
	
	return 1;
}

inline unsigned int
alloc_null_binding(struct ip_conntrack *conntrack,
		   struct ip_nat_info *info,
		   unsigned int hooknum)
{
	/* Force range to this IP; let proto decide mapping for
	   per-proto parts (hence not IP_NAT_RANGE_PROTO_SPECIFIED).
	   Use reply in case it's already been mangled (eg local packet).
	*/
	u_int32_t ip
		= (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC
		   ? conntrack->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip
		   : conntrack->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip);
	struct ip_nat_range range
		= { IP_NAT_RANGE_MAP_IPS, ip, ip, { 0 }, { 0 } };

	DEBUGP("Allocating NULL binding for %p (%u.%u.%u.%u)\n", conntrack,
	       NIPQUAD(ip));
	return ip_nat_setup_info(conntrack, &range, hooknum);
}

unsigned int
alloc_null_binding_confirmed(struct ip_conntrack *conntrack,
                             struct ip_nat_info *info,
                             unsigned int hooknum)
{
	u_int32_t ip
		= (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC
		   ? conntrack->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip
		   : conntrack->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip);
	u_int16_t all
		= (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC
		   ? conntrack->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u.all
		   : conntrack->tuplehash[IP_CT_DIR_REPLY].tuple.src.u.all);
	struct ip_nat_range range
		= { IP_NAT_RANGE_MAP_IPS, ip, ip, { all }, { all } };

	DEBUGP(KERN_DEBUG "Allocating NULL binding for confirmed %p (%u.%u.%u.%u)\n",
	       conntrack, NIPQUAD(ip));
	return ip_nat_setup_info(conntrack, &range, hooknum);
}

int ip_nat_rule_find(struct sk_buff **pskb,
		     unsigned int hooknum,
		     const struct net_device *in,
		     const struct net_device *out,
		     struct ip_conntrack *ct,
		     struct ip_nat_info *info)
{
	int ret;

	ret = ipt_do_table(pskb, hooknum, in, out, &nat_table, NULL);

	if (ret == NF_ACCEPT) {
		if (!ip_nat_initialized(ct, HOOK2MANIP(hooknum)))
			/* NUL mapping */
			ret = alloc_null_binding(ct, info, hooknum);
	}
	return ret;
}
#ifdef TEAMF1_CHANGES
static inline int
device_cmp(struct ip_conntrack *i, void *ifindex)
{
	int ret;

	read_lock_bh(&snat_lock);
	ret = (i->nat.snat_index == (int)(long)ifindex);
	read_unlock_bh(&snat_lock);

	return ret;
}
static inline int
sip_port_cleanup(struct ip_conntrack *i, void *port)
    {
	int ret;
    /*13C4 == 5060*/
	read_lock_bh(&snat_lock);
    ret = (i->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u.udp.port == SIP_PORT);
	read_unlock_bh(&snat_lock);
	return ret;

    }
static int snat_device_event(struct notifier_block *this,
			     unsigned long event,
			     void *ptr)
{
	struct net_device *dev = ptr;

	if (event == NETDEV_DOWN) {
		/* Device was downed.  Search entire table for
		   conntracks which were associated with that device,
		   and forget them. */
		IP_NF_ASSERT(dev->ifindex != 0);

		ip_ct_iterate_cleanup(device_cmp, (void *)(long)dev->ifindex);
	}
	if (event == NETDEV_GENERIC) {
        printk(KERN_DEBUG "%s: Cleaning sip entries\n",__func__);
        /*clean sip related conntracks*/
		ip_ct_iterate_cleanup(sip_port_cleanup, NULL);
    }

	return NOTIFY_DONE;
}

static int snat_inet_event(struct notifier_block *this,
			   unsigned long event,
			   void *ptr)
{
	struct net_device *dev = ((struct in_ifaddr *)ptr)->ifa_dev->dev;

	if (event == NETDEV_DOWN) {
		/* IP address was deleted.  Search entire table for
		   conntracks which were associated with that device,
		   and forget them. */
		IP_NF_ASSERT(dev->ifindex != 0);

		ip_ct_iterate_cleanup(device_cmp, (void *)(long)dev->ifindex);
	}

	return NOTIFY_DONE;
}

static struct notifier_block snat_dev_notifier = {
	.notifier_call	= snat_device_event,
};

static struct notifier_block snat_inet_notifier = {
	.notifier_call	= snat_inet_event,
};
#endif /* TEAMF1_CHANGES */

static struct ipt_target ipt_snat_reg = {
	.name		= "SNAT",
	.target		= ipt_snat_target,
	.checkentry	= ipt_snat_checkentry,
};

static struct ipt_target ipt_dnat_reg = {
	.name		= "DNAT",
	.target		= ipt_dnat_target,
	.checkentry	= ipt_dnat_checkentry,
};

#ifdef TEAMF1_OTO_IB_NAT_CHANGES
/*  creates /proc/sys/tf1/strictDnat, to enable/disable OTO IB NAT. */
static ctl_table strict_dnat_table[] = {
    { .ctl_name     = TF1_DBG + 2,  
      .procname     = "strictDnat",
      .data         = &strictDnat,
      .maxlen       = sizeof(strictDnat),
      .mode         = 0644,
      .proc_handler = &proc_dointvec 
    },
    {}
};

/*  This adds a child to /proc/sys/tf1 */
static ctl_table strict_dnat_root_table[] = {
    { .ctl_name    = TF1_DBG,
      .procname    = "tf1",
      .mode        = 0555,
      .child       = strict_dnat_table },
    {}
};
#endif

int __init ip_nat_rule_init(void)
{
	int ret;

	ret = ipt_register_table(&nat_table, &nat_initial_table.repl);
	if (ret != 0)
		return ret;
	ret = ipt_register_target(&ipt_snat_reg);
	if (ret != 0)
		goto unregister_table;

	ret = ipt_register_target(&ipt_dnat_reg);
	if (ret != 0)
		goto unregister_snat;
#ifdef TEAMF1_CHANGES    
	/* Register for device down reports */
    register_netdevice_notifier(&snat_dev_notifier);

	/* Register IP address change reports */
	register_inetaddr_notifier(&snat_inet_notifier);
#endif /* TEAMF1_CHANGES */
#ifdef TEAMF1_OTO_IB_NAT_CHANGES
    register_sysctl_table(strict_dnat_root_table, 1); 
#endif /* TEAMF1_OTO_IB_NAT_CHANGES */
	return ret;

 unregister_snat:
	ipt_unregister_target(&ipt_snat_reg);
 unregister_table:
	ipt_unregister_table(&nat_table);

	return ret;
}

void ip_nat_rule_cleanup(void)
{
	ipt_unregister_target(&ipt_dnat_reg);
	ipt_unregister_target(&ipt_snat_reg);
	ipt_unregister_table(&nat_table);
#ifdef TEAMF1_CHANGES    
	unregister_netdevice_notifier(&snat_dev_notifier);
	unregister_inetaddr_notifier(&snat_inet_notifier);    
#endif /* TEAMF1_CHANGES */    
#ifdef TEAMF1_OTO_IB_NAT_CHANGES
    unregister_sysctl_table(strict_dnat_root_table);
#endif /* TEAMF1_OTO_IB_NAT_CHANGES */
}
